from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import map
from builtins import zip
from builtins import str
from past.builtins import basestring
import pickle as pickle
from functools import wraps
import glob
import gzip
import json
import os
import pip
import time
import urllib.request, urllib.parse, urllib.error
import shutil
import zipfile
from zipfile import ZIP_DEFLATED, ZipFile

import boto3
import boto3.session
from threading import Thread

from botocore.exceptions import ClientError
from cloud_resources import storage_catalyst
from db_samvaad.db_handler import DBMethod
import pandas as pd
import requests

from algowrap import cloudfont_utils as cf
from algowrap import constants as cnsts
from algowrap import error_handler as err_hndl
from algowrap.sentry_client import sentry_client
from algowrap import credentials as creds
from algowrap import err_constants as err_cnsts
from algowrap import shared_vars




def folder_appender(file_object, folder_name):
    """ This function appends the folder name to the start
        of the file
        Inputs: file_object : which will have all the file names
                folder_name : folder name that has to be appened to
                the files
    """
    return_obj = dict()
    for key, value in list(file_object.items()):
        if isinstance(value, list):
            return_obj[key] = list()
            for val in value:
                return_obj[key].append(str(folder_name) + os.sep + val)
        else:
            return_obj[key] = str(folder_name) + os.sep + value

    return return_obj


def wildcard_present(file_name):
    return "*" in file_name.get('path')


def get_iteration_output_key(iteration_id, parameter_json, upload_user_bucket=False):
    """This function returns the parameter file json
    Input: iteration id
    Return: heartbeat which will be a json object which
    contains the file name
    """
    # TODO: file location that is being uploaded is now the same
    # as download we would have to change it
    iter_type = DBMethod.get_value_from_iter_id(iteration_id, 'type')

    if iter_type == cnsts.ITER_WORKFLOW:
        parent_id = DBMethod.get_value_from_iter_id(iteration_id,
                                                     cnsts.WORKFLOW_REQUEST)
    elif iter_type == cnsts.ITER_COMPONENT:
        parent_id = DBMethod.get_value_from_iter_id(iteration_id,
                                                     cnsts.RUN_IDFR)
    else:
        raise AttributeError(err_cnsts.ITERERR1_INVALID_ITER)
    
    if(shared_vars.UPLOAD_USER_BUCKET and upload_user_bucket):
        folder_append = str(parent_id)
    else:
        folder_append = str(parent_id) + os.sep + str(iteration_id)

    coloum_name = cnsts.DOWNLOAD_JSON_FILE
    parameter_json = DBMethod.get_value_from_iter_id(iteration_id, coloum_name)

    files_s3 = folder_appender(parameter_json, folder_append)

    # result_data = dict()

    # for file_out in files_s3:
    #     result_data[file_out] = cf.get_iterationoutput_signed_url(
    #                                         files_s3[file_out])

    return files_s3


def get_iteration_input_signed_url(parameter_data):
    """This function returns the signed url of input that is needed for
        the algorithm wrapper to to download the input files
        Input: iteration id
        Returns: the urls that is needed to process the output
    """
    if cnsts.USER_INPUT_IDENTIFIER not in parameter_data:
        raise err_hndl.InvalidInputError(err_cnsts.INPUTERR13_INVALID_PARAMETER_FILE)
    else:
        input_files = parameter_data.get(cnsts.USER_INPUT_IDENTIFIER)

    pre_signed_urls = dict()
    for file_key, file_name in list(input_files.items()):
        if(isinstance(file_name, list)):
            for fil_nam in file_name:
                if file_key not in pre_signed_urls:
                    pre_signed_urls[file_key] = []

                s3_bucket = fil_nam.get('bucket')
                s3_key = fil_nam.get('path')

                pre_signedurl = cf.generate_presigned_url(s3_key, s3_bucket)
                pre_signed_urls[file_key].append({'url': pre_signedurl,
                                                  'file_name': fil_nam.get('path'),
                                                  'bucket': s3_bucket
                                                  })
        else:
            s3_bucket = file_name.get('bucket')
            s3_key = file_name.get('path')

            pre_signedurl = cf.generate_presigned_url(s3_key, s3_bucket)
            pre_signed_urls[file_key] = {'url': pre_signedurl,
                                         'file_name': file_name.get('path'),
                                         'bucket': s3_bucket
                                         }

    return {'urls': pre_signed_urls}


def log_tracking_packages():
    """
    Add required python package versions to sentry debug log.

    Downloads(from s3) list of python packages to be logged and adds version
    of every package in that list to the debug log (using pip). It can
    thus be a part of Sentry logs. Example:
        algowrap 0.1.30
        sentrylogger 1.5.5
                        if the file containing list of packags to be logged
                        contains only 'algowrap' and 'sentrylogger' separated
                        by a newline character.
    """
    try:
        package_list_s3 = requests.get(cnsts.LOG_PACKAGES_URL)
        package_list = []
        dwnlded_list = [y for y in package_list_s3.text.split('\n') if y != '']
        for pack in dwnlded_list:
            package_list.append(pack)
        sentry_client.queue_debug_log(cnsts.LIST_PACKGES)
        for pack in map(str, pip.get_installed_distributions(local_only=True)):
            for privatepack in package_list:
                if privatepack in pack:
                    sentry_client.queue_debug_log(str(pack))
    except (Exception):
        sentry_client.queue_debug_log(cnsts.LOG_PACK_S3_ERR)


def url_wild_card_replacer(url, replacee):
    """
    Returns a url from a wild card url and a list of replacers.

    Checks if length of wild characters (*) matches the number of replacers.
    Then, replaces every * with the ordered replacer.

    Parameters
    ----------
    url: String
        A url with wild characters.
    replacee: List
        Ordered items to replace wild characters with.

    Returns
    -------
    String: new_url
        A URL without wild characters, generated from wild card url.
    """
    number_stars = url.count("*")
    star_split = url.split('*')
    new_url = ""
    if number_stars == len(replacee):
        for s, r in zip(star_split, replacee):
            new_url += s + r
        new_url += star_split[len(star_split) - 1]
    else:
        raise err_hndl.InvalidInputError(err_cnsts.URLERR1_URL_WILDCARD_REPL)
    return new_url


def split_file_upload_hlpr(output_dataobject, directory_name, output_key,
                           output_data):
    """
    Helper function for uploading split file data.

    Data is split when it is large/long. This function creates a temporary file
    on local system and uploads the split data by converting to json using
    to_json on DataFrames. Then uses cUrl PUT command to upload the data with
    the help of wild card urls.

    Parameters
    ----------
    output_dataobject: Dictionary
        Contains information for the file type and data to upload.
    directory_name: String
        Name of local directory to use for temporary files.
    output_key: String
        Contains information regarding file type to upload.
    output_data: String
        s3 URL to upload the data.
    """
    
    file_var = 0
    for split_data in output_dataobject[output_key]:
        split_path_local = directory_name + cnsts.SEPARAT + 'result' + str(file_var) + '.json'
        file_var = file_var + 1
        if isinstance(split_data.get('data'), pd.DataFrame):
            split_data.get('data').to_json(split_path_local,
                                           orient='records')
        elif not (isinstance(split_data.get('data'), pd.DataFrame))\
                and not (isinstance(split_data.get('data'), dict)) \
                and isinstance(split_data.get('data'), basestring)\
                and os.path.exists(split_data.get('data')):
            split_path_local = split_data.get('data')
        else:
            with open(split_path_local, 'w') as outfile:
                outfile.write(json.dumps(split_data.get('data')))

        thread = Thread(target=upload_to_s3, args=(creds.ITER_OUTPUT_S3_BUCKET, url_wild_card_replacer(output_data, split_data.get('name_filler')),split_path_local))
        thread.start()
        thread.join()

def pickle_upload_hlpr(output_dataobject, directory_name, output_key,
                       output_data):
    """
    Helper function for writing serialized data to local directory and then
    uploading to s3.

    cPickle is a serializer package for python. The serialized data here is
    referred to as pickle. Generally, pickle or pickling refers to the process
    of serializing the data.

    Parameters
    ----------
    directory_name: String
        Local directory name to write temporary file.
    output_dataobject: Dict
        Contains output data.
    output_key: String
        The key for which data is to be extracted from the Dictionary.
    output_data: String
        s3 URL to upload the data.
    """
    file_path_local = directory_name + cnsts.SEPARAT + 'result.json'
    with open(file_path_local, 'wb') as handle:
        pickle.dump(output_dataobject[output_key],
                    handle, protocol=pickle.HIGHEST_PROTOCOL)
    upload_to_s3(creds.ITER_OUTPUT_S3_BUCKET,output_data,file_path_local)


def compressed_upload_hlpr(output_dataobject, directory_name, output_key,
                           output_data):
    """
    Helper function for writing compressed data to local directory and
    uploading to s3.

    Writes to a .gz file using gzip python package.
    Parameters
    ----------
    directory_name: String
        Local directory name to write temporary file.
    output_dataobject: Dict
        Contains output data.
    output_key: String
        The key for which data is to be extracted from the Dictionary.
    output_data: String
        s3 URL to upload the data.
    """
    file_path_local = directory_name + cnsts.SEPARAT + 'result.json.gz'
    with gzip.open(file_path_local, 'wb') as handle:
        handle.write(output_dataobject[output_key])

    upload_to_s3(creds.ITER_OUTPUT_S3_BUCKET,output_data,file_path_local)


def rds_upload_hlpr(output_dataobject, directory_name, output_key,
                    output_data):
    """
    Helper function for writing RDS data to local directory and uploading to s3

    Opens a file using .rds extension and simply writes to it. Then calls the
    cURL upload helper function to make a cURL to upload to s3.

    Parameters
    ----------
    directory_name: String
        Name of local directory to create a temporary result file on.
    output_dataobject: Dict
        Contains output data.
    output_key: String
        The key for which data is to be extracted from the Dictionary.
    output_data: String
        s3 URL to upload the data.
    """
    file_path_local = directory_name + cnsts.SEPARAT + 'result.rds'
    with open(file_path_local, 'wb') as outfile:
        outfile.write(output_dataobject[output_key])

    upload_to_s3(creds.ITER_OUTPUT_S3_BUCKET,output_data,file_path_local)


def json_upload_hlpr(output_dataobject, directory_name, output_key,
                     output_data):
    """
    Helper function for writing JSON data to local directory and uploading to
    s3.

    Parameters
    ----------
    file_path_local: String
        Local file path to write data to.
    output_dataobject: Dict
        Contains output data.
    output_key: String
        The key for which data is to be extracted from the Dictionary.
    output_data: String
        s3 URL to upload the data.
    """
    file_path_local = directory_name + cnsts.SEPARAT + 'result.json'
    with open(file_path_local, 'w') as outfile:
        outfile.write(json.dumps(output_dataobject[output_key]))
    
    upload_to_s3(creds.ITER_OUTPUT_S3_BUCKET,output_data,file_path_local)


def upload_result_files(output_file_types, output_dataobject, directory_name, iter_id, upload_user_bucket=False):
    """
    This function uploads the result files to S3.

    Handles different types of output files and uploads them accordingly.
    Writes a temporary result file to given local directory_name and uploads to
    s3. Uses cUrl PUT command for uploading.

    Parameters
    ----------
    output_file_types: Dictionary
        Contains information regarding output file types and URL.
    output_dataobject: Dictionary
        Contains information regarding the output file types and data object to
        write.
    directory_name: String
        Name of local directory to write temporary result files on.
    iter_id : integer
        Iteration Id
    """

    if(shared_vars.UPLOAD_USER_BUCKET and upload_user_bucket):
        output_bucket = creds.USER_INPUT_S3_BUCKET
        if not isinstance(output_dataobject, list):
            #check for list of files to be uploaded to input bucket
            ''' 
            DICT: output_dataobject
            KEY: 'Upload_to_input' : It contains list of files that are to be uploaded to S3.
            Example: output_dataobject{'Upload_to_input' : ['raw_input_files', 
                'maven_output_peaks', 'raw_input_file']}
            '''
            if cnsts.UPLOAD_INPUT in output_dataobject:
                copy_output_file_types = output_file_types.copy()
                output_file_types.clear()
                list_input_filenames = output_dataobject[cnsts.UPLOAD_INPUT]
                for filename in list_input_filenames:
                    if filename in copy_output_file_types:
                        #add keys for files to be uploaded to input bucket
                        output_file_types[filename] = copy_output_file_types[filename]
    else:
        output_bucket = creds.ITER_OUTPUT_S3_BUCKET
    for output_key, output_data in list(output_file_types.items()):
        if not isinstance(output_dataobject, list):
            if output_key in output_dataobject:
                if isinstance(output_dataobject[output_key], pd.DataFrame):
                    file_path_local = directory_name + cnsts.SEPARAT + 'result.json'
                    output_dataobject[output_key].to_json(file_path_local,
                                                        orient='records')
                    upload_to_s3(output_bucket,output_data,file_path_local)
                elif not(isinstance(output_dataobject[output_key], pd.DataFrame)) \
                        and not(isinstance(output_dataobject[output_key], dict)) \
                        and isinstance(output_dataobject[output_key], basestring) \
                        and os.path.exists(output_dataobject[output_key]):
                    upload_to_s3(output_bucket,output_data,output_dataobject[output_key])
                else:
                    if 'pickle' in output_key:
                        pickle_upload_hlpr(output_dataobject, directory_name,
                                        output_key, output_data)
                    elif 'compressed' in output_key:
                        compressed_upload_hlpr(output_dataobject, directory_name,
                                            output_key, output_data)
                    elif 'split' in output_key:
                        split_file_upload_hlpr(output_dataobject, directory_name,
                                            output_key, output_data)
                    elif 'rds' in output_key:
                        rds_upload_hlpr(output_dataobject, directory_name,
                                        output_key, output_data)
                    elif cnsts.PATH_IDFR in output_key:
                        upload_to_s3(creds.ITER_OUTPUT_S3_BUCKET,output_data,output_dataobject[output_key])
                    else:
                        json_upload_hlpr(output_dataobject, directory_name,
                                        output_key, output_data)
        else:
            upload_files_acc_to_format(output_dataobject, output_key, directory_name,
                                    output_data, iter_id)


def upload_files_acc_to_format(output_dataobject, output_key, directory_name,
                                    output_data, iter_id):
    """
    This function check format key which are comming from the algorithm
    for each file and call different function according to that 
    to upload the file/folder to S3

    directory_name: String
        Local directory name to write temporary file.
    output_dataobject: list of object
        each object contain two key values
        ex- {"file_name" : file_name, "format" : "zip/unzip"}
    output_key: String
        The key for which data is to be extracted from the Dictionary
    output_data: String
        S3 file path which is getting from algorithm csv file
    """

    # Iterting over the list which contain objects and each object
    # contain two keys, one is file name and 2nd one is format
    for each_file in output_dataobject:

        # There are only two keys, one is file name and 2nd one is format
        if output_key in each_file:

            # Format maybe zip or unzip
            format = each_file[cnsts.OUTPUT_FORMAT]
            if format == cnsts.ZIP:
                upload_zip_folder_to_s3(each_file[output_key], directory_name,
                    output_key, output_data, iter_id)
            elif format == cnsts.UNZIP:
                op_folder_upload_hlpr(each_file[output_key], output_data)
        

def upload_zip_folder_to_s3(op_file_path, directory_name, output_key,
                               output_data, iter_id):
    """
    Helper function for writing compressed zip data to local directory and
    uploading to s3.

    Writes to a .zip file using Zipfile python library.

    Parameters
    ----------

    op_file_path : str
        output folder path
    directory_name: String
        Local directory name to write temporary file.
    output_dataobject: Dict
        Contains output data.
    output_key: String
        The key for which data is to be extracted from the Dictionary.
    output_data: String
        output file name which is getting from the algorithm csv
    iter_id : String
        Iteration id
    """

    # Output zip file path where zip file would be create
    iter_id = op_file_path.split('/')[3]
    op_zip_file_path = creds.MOUNT_FILE_PATH + os.sep + iter_id + os.sep + output_key

    # Using ZipFile to make zip
    zf = zipfile.ZipFile("%s.zip" % (op_zip_file_path), "w", zipfile.ZIP_DEFLATED)
    abs_src = os.path.abspath(op_file_path)

    # Using this format So that zip doesn't contain the root folder
    for dirname, subdirs, files in os.walk(op_file_path):
        for filename in files:
            absname = os.path.abspath(os.path.join(dirname, filename))
            arcname = absname[len(abs_src) + 1:]
            zf.write(absname, arcname)
    zf.close()

    # Calling functio to upload zip file to S3
    run_id = output_data.split("/")[0]
    zip_file_name = output_data.split("/")[-1]
    s3_key_path = str(run_id) + os.sep + str(iter_id) + os.sep + zip_file_name
    upload_to_s3(creds.ITER_OUTPUT_S3_BUCKET, s3_key_path,
                                op_zip_file_path + cnsts.ZIP_EXTENSION)
                        

def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=None):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :type ExceptionToCheck: Exception or tuple
    :param tries: number of times to try (not retry) before giving up
    :type tries: int
    :param delay: initial delay between retries in seconds
    :type delay: int
    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    :type backoff: int
    :param logger: logger to use. If None, print
    :type logger: logging.Logger instance
    """
    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    msg = "%s, Retrying in %d seconds..." % (str(e), mdelay)
                    if logger:
                        logger.warning(msg)
                    else:
                        print(msg)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry


def download_save(s3_key, bucket,download_path):
    """
    Download data from the given s3 bucket to save to a given local path.

    Parameters
    ----------
    s3_key: str
        s3 key which has to be downloaded
    bucket : str
        bucket from which the files needs to be downloaded
    local_file_save_path: str
        Local file path to save downloaded data.
    """
    try:
        client = boto3.client('s3')
        client.download_file(bucket, s3_key,download_path)
    except ClientError as error:
        raise err_hndl.S3OperationError(str(error))


def upload_to_s3(bucket, s3_key, local_file):
    '''
    upload file to s3  from the given local path

    PARAMETERS
    ----------
    bucket : str
        bucket to which the file needs to be uploaded
    s3_key : str
        s3 key where the file has to be uploaded
    local_file : str
        Local file path where the file is present locally
    '''
    try:
        s3_client = boto3.client('s3')
        s3_client.upload_file(local_file, bucket, s3_key)
    except ClientError as error:
        raise err_hndl.S3OperationError(str(error))



def get_param_json(iter_id):
    """
    Returns data from parameter.json file of a particular iteration.

    Checks for the iteration type for a given iteration and fetches the param.json
    corresponding to it.

    Parameters
    ----------
    iter_id: String
        iteration id to get parameter.json for.

    Returns
    -------
    JSON: parameter_data
        Data from parameter.json
    """
    iter_type = DBMethod.get_value_from_iter_id(iter_id, 'type')
    s3_file_key = get_s3_key(iter_id, iter_type)
    parameter_data = json.loads(get_user_objet(s3_file_key, iter_type))
    return parameter_data


def get_s3_key(iter_id, iter_type):
    """
    Returns the s3 key for given iteration id from dataDBMethod.

    Checks if the corresponding iteration is a workflow request or not and
    fetches the parameter file accordingly.

    Parameters
    -----------
    iter_id: Integer
        iteration id to fetch s3 key

    iter_type: String
        Type of iteration

    Returns
    -------
    String: s3_file_key
        S3 file key for given iteration id.
    """
    if iter_type == cnsts.ITER_WORKFLOW:
        parent_id = DBMethod.get_value_from_iter_id(iter_id, cnsts.WORKFLOW_REQUEST)
    elif iter_type == cnsts.ITER_COMPONENT:
        parent_id = DBMethod.get_value_from_iter_id(iter_id,cnsts.RUN_IDFR )
    else:
        raise AttributeError(err_cnsts.ITERERR1_INVALID_ITER)

    parameter_filename = DBMethod.get_value_from_iter_id(iter_id, cnsts.PARAM_FILE)
    parameter_file = parameter_filename[cnsts.PARAM_KEY]
    s3_file_key = str(parent_id) + os.sep + str(iter_id) + os.sep + parameter_file
    return s3_file_key



def get_user_objet(s3_key, iter_type):
    """
    This function gives the s3 object from the bucket mithoo-user-input
    Inputs: full path to the s3 object
    Returns : The S3 object in the form of string
    """

    bucket_name = creds.USER_INPUT_S3_BUCKET

    return storage_catalyst.get_object(s3_key, bucket_name)


def op_folder_upload_hlpr(op_file_path, output_data):
    """
    This function upload the output folder or files to S3 by checking
    it's directory

    Parameters
    ------------
    output_data : str
        output file nama which is getting from the algorithm csv
    op_file_path : str
        this is output file/folder path
    """

    # getting run id
    run_id = output_data.split("/")[0]

    # checking output which have to push to s3 is a folder or file
    if not os.path.isdir(op_file_path):
        upload_to_s3(creds.ITER_OUTPUT_S3_BUCKET, output_data, op_file_path)
    else:
        # iterating over a output folder which contain many files and
        # format can be anything
        for folder, dirs, files in os.walk(op_file_path):
            op_files = glob.glob(folder + "/*.*")

            # uploading all the files to S3 using thread
            for each_file in op_files:
                thread = Thread(target = upload_to_s3, args=(creds.ITER_OUTPUT_S3_BUCKET,
                    each_file.replace(creds.MOUNT_FILE_PATH, run_id), each_file))
                thread.start()
